from insightspy.SessionCore import RequestCore
from insightspy.utils import _list_to_int_list
import pandas as pd


class ComparisonsMixin(RequestCore):
    def comparisons(self):
        """List comparisons

        Lists comparisons accessible within the current portal session. Access is
        limited by user id.

        Examples:
            >>> # p is a logged in portal session
            >>> p.comparisons()

        Returns:
            DataFrame: comparison metadata
        """
        export_keys = [
            "comparison_id",
            "description",
            "genome",
            "control_label",
            "treatment_label",
            "treatment_ids",
            "control_ids",
        ]
        data = self._post("comparisons/retrieve", expect_data=True)["response"]["data"]
        if data is None:
            return pd.DataFrame()
        out = [{key: v[key] for key in export_keys} for v in data]
        out = pd.DataFrame.from_dict(out)
        return out

    def get_comparisons(self, comparison_ids):
        """List comparisons

        Get comparison results

        Args:
            comparison_ids (int, [int]):  one or more sample ids

        Examples:
            >>> # p is a logged in portal session
            >>> p.get_comparisons([4,8])

        Returns:
            DataFrame:
                comparison results from DESeq2. Refer to DESeq2 documentation for column
                description
        """
        if not isinstance(comparison_ids, list):
            comparison_ids = [comparison_ids]
        comparison_ids = _list_to_int_list(comparison_ids)
        data = self._post(
            "comparisons/retrieve",
            {"comparison_ids": comparison_ids, "withData": True},
            expect_data=True,
        )["response"]["data"]
        if len(data) > 0:
            out = {
                str(i): g.drop(columns="comparison_id")
                for i, g in pd.DataFrame.from_dict(data).groupby("comparison_id")
            }
        else:
            out = {}
        return out



from insightspy.SessionCore import RequestCore
from insightspy.utils import _list_to_int_list, _parse_locus
import pandas as pd


class SignaturesMixin(RequestCore):
    def signatures(self):
        """List signatures

        Lists signatures accessible within the current portal session. Access is
        limited by user id.

        Examples:
            >>> # p is a logged in portal session
            >>> p.signatures()

        Returns:
            DataFrame: signature metadata
        """
        export_keys = ["signature_id", "description", "annotation_id", "size"]
        data = self._post("signature/retrieve", {"noData": True}, expect_data=True)[
            "response"
        ]["data"]
        if data is None:
            return pd.DataFrame()
        out = pd.DataFrame.from_records(
            [{key: v[key] for key in export_keys} for v in data.values()]
        )
        out.rename(columns={"annotation_id": "parent_reference"}, inplace=True)
        return out

    def get_signatures(self, signature_ids):
        """Get signature set

        Get signature set coordinates as a pandas dataframe

        Args:
            signature_id (int): a signature_id

        Examples:
            >>> # p is a logged in portal session
            >>> p.get_signatures(8)

        Returns:
            {DataFrame}:
                dictionary of DataFrames containing the chromosomal co-ordinates in (,]
                intervals for the signatures. The dictionary keys are the signature ids
        """
        if not isinstance(signature_ids, list):
            signature_ids = [signature_ids]
        signature_ids = _list_to_int_list(signature_ids)
        d = self._post(
            "signature/retrieve",
            {"signature_ids": signature_ids, "noData": False},
            expect_data=True,
        )["response"]["data"]
        colnames = ["seqname", "start", "end", "name"]
        out = {
            k: pd.DataFrame.from_records(
                [
                    _parse_locus(x["location"]) + [x["feature_id"]]
                    for x in v["features"]
                ],
                columns=colnames,
            )
            for k, v in d.items()
        }
        return out



from insightspy.SessionCore import RequestCore
import pandas as pd


class PipelinesMixin(RequestCore):
    def pipeline_revisions(self):
        """List available pipelines

        Lists pipelines accessible within the current portal session. Pipeline access is
        limited by user id and the current session project if one is set.

        Examples:
            >>> # p is a logged in portal session
            >>> p.pipeline_revisions()

        Returns:
            DataFrame: table of pipeline metadata
        """
        return pd.DataFrame.from_dict(
            self._get("pipelinesDb/project_pipelines_revisions")["response"]["data"]
        )



from insightspy.ProjectsMixin import ProjectsMixin
from insightspy.SamplesMixin import SamplesMixin
from insightspy.PipelinesMixin import PipelinesMixin
from insightspy.ReferenceSetMixin import ReferenceSetsMixin
from insightspy.ComparisonsMixin import ComparisonsMixin
from insightspy.SignaturesMixin import SignaturesMixin
import requests
import getpass


class PortalSession(
    ProjectsMixin,
    SamplesMixin,
    PipelinesMixin,
    ReferenceSetsMixin,
    ComparisonsMixin,
    SignaturesMixin,
):
    def __init__(self, api_key=None, url="https://insights.arpeggiobio.com"):
        session = requests.Session()
        self._session = session
        self.url = url
        if api_key is not None:
            self._update_credentials(api_key)

    def login(self, email=None, password=None, api_key=None):
        """Login to Arpeggio portal

        Logs in current session to the Arpeggio portal. All subsequent requests will be
        authenticated with these credentials. Will try the `api_key` first if
        specified. If `api_key` is not present can use the same `email` and `password`
        that is used to log in to the portal. If email is entered without as password
        the user will be asked to enter their password. NOTE: we do not reccommend
        saving your password in a script, either access it as an environment variable or
        enter it interactively.

        Args:
            api_key (str): login key
            email (str): email used to login to the arpeggio portal
            password (str): password used to login to the arpeggio portal

        Examples:
            >>> p = PortalSession()
            >>> p.login(email = user@domain.com)
        """
        if api_key is None:
            if email is not None and password is None:
                password = getpass.getpass()
            elif email is None and password is None:
                raise ValueError("Please specify login credentials")
            access_response = self._post(
                "user/login", {"email": email, "password": password}, True
            )
            self._update_credentials(access_response["response"]["data"]["token"])
        else:
            raise NotImplementedError("API key login not yet implemented")
        print("Login succeeded")



from insightspy.SessionCore import RequestCore
from insightspy.utils import _list_to_int_list, _download
import pandas as pd
import os


class SamplesMixin(RequestCore):
    def samples(self, tag_suffix="_tag_dupl", sample_ids=None):
        """List samples

        Lists samples accessible withthin the current portal session. Sample access is
        limited by user id and the current session project if one is set.

        Args:
            tag_suffix (str):  the suffix that is appended to the column name if a tag
                shares the same name as one of the columns in the core database used to
                describe the samples
            sample_ids (int, [int]):  one or more sample ids to filter on (optional).
                This will also override any filtering from the project scope.

        Examples:
            >>> # p is a logged in portal session
            >>> p.samples()
            >>> # To only retrieve metadata about specific samples
            >>> p.samples(sample_ids = [32, 35, 42])

        Returns:
            DataFrame: table of sample metadata
        """
        if sample_ids is not None:
            ignore_project = True
            if not isinstance(sample_ids, list):
                sample_ids = [sample_ids]
            sample_ids = _list_to_int_list(sample_ids)
        else:
            ignore_project = False
        # List keys to export to user
        export_keys = [
            "sample_id",
            "description",
            "sample_type",
            "reference_genome",
            "revision_id",
            "tags",
        ]
        out = [
            {key: v[key] for key in export_keys}
            for v in self._post(
                "sample/retrieve",
                {
                    "tagsAsDict": True,
                    "sample_ids": sample_ids,
                    "ignoreProject": ignore_project,
                },
                expect_data=True,
            )["response"]["data"]
        ]
        out = pd.DataFrame.from_dict(out)
        if "tags" in out:
            out = out.join(
                pd.DataFrame(out.pop("tags").values.tolist()), rsuffix=tag_suffix
            )
        return out

    def gene_tpm(self, sample_ids):
        """Get gene TPMs

        Get transcripts per million (TPMs) values for all pre-quantified genes in a
            sample

        Args:
            sample_ids (int, [int]):  one or more sample ids

        Examples:
            >>> # p is a logged in portal session
            >>> p.gene_tpm([33, 48])

        Returns:
            DataFrame:
                table of gene TPMs with sample ids as column names and gene names as row
                names
        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        counts = self._post(
            "sample/TPM",
            {
                "sample_ids": sample_ids,
                "minimal": True,
                "raw_counts": False,
                "pseudocounts": 0,
            },
            expect_data=True,
        )["response"]["data"]
        out = pd.DataFrame.from_dict(counts[0], orient="index").sort_index()
        out.columns = [str(x["sample_id"]) for x in counts[1]]
        return out

    def gene_count(self, sample_ids):
        """Get gene TPMs

        Get read counts for all pre-quantified genes in a sample

        Args:
            sample_ids (int, list):  one or more sample ids

        Examples:
            >>> # p is a logged in portal session
            >>> p.gene_count([33, 48])

        Returns:
            DataFrame:
                table of read counts with sample ids as column names and gene names
                as row names
        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        counts = self._post(
            "sample/TPM",
            {
                "sample_ids": sample_ids,
                "minimal": True,
                "raw_counts": True,
                "pseudocounts": 0,
            },
            expect_data=True,
        )["response"]["data"]
        out = pd.DataFrame.from_dict(counts[0], orient="index").sort_index()
        out.columns = [str(x["sample_id"]) for x in counts[1]]
        return out

    def locus_density(self, sample_ids, locus):
        """Get locus density profile

        Get binned coverage counts for a locus

        Args:
            sample_ids (int, [int]):  one or more sample ids
            locus(str): locus in the form "chr:start-stop"

        Examples:
            >>> # p is a logged in portal session
            >>> p.locus_density(33, "chr1:1000000-1100000")

        Returns:
            dict(DataFrame):
                dictionary of DataFrames containing RPK normalized mean signal
                per bin with one value per strand. One entry in the dictionary
                per sample being queried.
        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        density = self._post(
            "density/locus",
            {"sample_ids": sample_ids, "locus": locus},
            expect_data=True,
        )["response"]["data"]
        out = {}
        # Iterate over all returned values (2 per sample, one for each strand) and
        # place them in a dictionary of DataFrames where the key is the sample id
        for l in density:
            id = str(l["ID"])
            x = pd.DataFrame.from_dict(l["values"])
            x["strand"] = l["strand"]
            if id not in out:
                out[id] = x
            else:
                out[id] = pd.concat([out[id], x])
        for k in out.keys():
            out[k].columns = ["bin_center", "value", "strand"]
        return out

    def download_fastq(self, sample_ids, target_directory="."):
        """Download fastq

        Download the fastq file(s) associated with the sample id(s)

        Args:
            sample_ids (int, [int]):  one or more sample ids
            target_directory(str): path to directory where files should be downloaded
                to. Will be created if it does not exist

        Examples:
            >>> # p is a logged in portal session
            >>> p.download_fastq(33, "./my_project/fastq")

        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        os.makedirs(target_directory, exist_ok=True)
        types = ["fastq", "fastq2"]
        urls = {}
        for i in sample_ids:
            for t in types:
                filesize = self._post(f"fileSize/sample/{t}/{i}")["response"]["data"]
                if filesize > 0:
                    read_suffix = "R1" if t == "fastq" else "R2"
                    filepath = os.path.join(
                        target_directory, f"{i}_{read_suffix}.fastq.gz"
                    )
                    urls[filepath] = self._post(f"download/sample/{t}/{i}")["response"][
                        "data"
                    ]
        _download(urls)

    def download_bam(self, sample_ids, target_directory="."):
        """Download bam

        Download the bam file(s) associated with the sample id(s)

        Args:
            sample_ids (int, [int]):  one or more sample ids
            target_directory(str): path to directory where files should be downloaded
                to. Will be created if it does not exist

        Examples:
            >>> # p is a logged in portal session
            >>> p.download_bam(33, "./my_project/bam")

        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        os.makedirs(target_directory, exist_ok=True)
        urls = {}
        for i in sample_ids:
            filesize = self._post(f"fileSize/sample/bam/{i}")["response"]["data"]
            if filesize > 0:
                filepath = os.path.join(target_directory, f"{i}.bam")
                urls[filepath] = self._post(f"download/sample/bam/{i}")["response"][
                    "data"
                ]
        _download(urls)

    def download_bigwig(self, sample_ids, target_directory="."):
        """Download bam

        Download the bigwig file(s) associated with the sample id(s)

        Args:
            sample_ids (int, [int]):  one or more sample ids
            target_directory(str): path to directory where files should be downloaded
                to. Will be created if it does not exist

        Examples:
            >>> # p is a logged in portal session
            >>> p.download_bigwig(33, "./my_project/bigwig")

        """
        if not isinstance(sample_ids, list):
            sample_ids = [sample_ids]
        sample_ids = _list_to_int_list(sample_ids)
        os.makedirs(target_directory, exist_ok=True)
        # Dictionary of the file types to use for both stranded and unstranded data
        suffix_dict = {
            "stranded": {
                "bigwigPos": "pos.normalized.bw",
                "bigwigNeg": "neg.normalized.bw",
            },
            "unstranded": {"bigwig": "normalized.bw"},
        }
        # set of the stranded datatypes
        stranded_datatypes = set(["proseq", "PRO-seq", "rnaseq", "RNA-seq"])
        # get sample info
        sample_info = self.samples(sample_ids=sample_ids).set_index("sample_id")
        urls = {}
        for i in sample_ids:
            filesize = self._post(f"fileSize/sample/bigwig/{i}")["response"]["data"]
            is_stranded = sample_info.loc[i].sample_type in stranded_datatypes
            str_key = "stranded" if is_stranded else "unstranded"
            for t in suffix_dict[str_key].keys():
                if filesize > 0:
                    suffix = suffix_dict[str_key][t]
                    filepath = os.path.join(target_directory, f"{i}.{suffix}")
                    urls[filepath] = self._post(f"download/sample/{t}/{i}")["response"][
                        "data"
                    ]
        _download(urls)



from insightspy.SessionCore import RequestCore
from insightspy.utils import _parse_locus
import pandas as pd


class ReferenceSetsMixin(RequestCore):
    def reference_sets(self, tag_suffix="_tag_dupl"):
        """List reference sets

        Lists reference sets accessible withthin the current portal session. Project
            access is limited by user id.

        Args:
            tag_suffix (str):  the suffix that is appended to the column name if a tag
                shares the same name as one of the columns in the core database used to
                describe the reference sets

        Examples:
            >>> # p is a logged in portal session
            >>> p.reference_sets()

        Returns:
            DataFrame: table annotation metadata
        """
        export_keys = [
            "annotation_id",
            "description",
            "reference_genome",
            "size",
            "tags",
        ]
        out = [
            {key: v[key] for key in export_keys}
            for v in self._post(
                "annotation/retrieve", {"tagsAsDict": True}, expect_data=True
            )["response"]["data"]
        ]
        out = pd.DataFrame.from_dict(out)
        out.rename(columns={"annotation_id": "reference_id"}, inplace=True)
        return out.join(
            pd.DataFrame(out.pop("tags").values.tolist()), rsuffix=tag_suffix
        )

    def get_reference_set(self, reference_id):
        """Get reference set

        Get reference set as a pandas dataframe

        Args:
            reference_id (int): a reference_id

        Examples:
            >>> # p is a logged in portal session
            >>> p.get_reference_set(5)

        Returns:
            DataFrame: chromosomal co-ordinates in (,] intervals
        """
        if not isinstance(reference_id, (int, float)):
            raise TypeError("reference_id must be an integer")
        if int(reference_id) == 1:
            raise NotImplementedError(
                "Reference set 1 is a special set in the portal "
                + "whose export is not currently supported"
            )
        d = self._post(
            "annotation/features",
            {"annotation_id": int(reference_id)},
            expect_data=True,
        )["response"]["data"]
        out = pd.DataFrame.from_records(
            [
                _parse_locus(x["location"]) + [x["feature_id"]]
                for x in d[str(reference_id)]["ids"]
            ]
        )
        out.columns = ["seqname", "start", "end", "name"]
        return out



# Flatten the namespace
from .PortalSession import *  # NOQA
import insightspy.utils
import sys

# If in an interactive session
if hasattr(sys, "ps1"):
    # check for portal connection and if that succeeds check package version
    if not insightspy.utils._connect("https://insights.arpeggiobio.com"):
        if not insightspy.utils._connect("https://www.google.com/"):
            print(
                "You are not connected to the internet (or google is down) "
                + "so will be unable to access the Arpeggio Portal"
            )
        else:
            print(
                "Arpeggio portal is not responding. Contact Arpeggio Biosciences "
                + "if this persists."
            )
    else:
        insightspy.utils._versions("insightspy")



"""
__version__.py
~~~~~~~~~~~~~~
Information about the current version of the py-package-template package.
"""

__title__ = "insightspy"
__description__ = "Python API for Arpeggio Biosciences data portal"
__version__ = "0.0.4"
__author__ = "Noah Dukler"
__author_email__ = "noah@arpeggiobio.com"
__license__ = "MIT"
__url__ = "https://github.com/ArpeggioBiosciences/insightspy"
__classifiers__ = [
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
]



import re
import json
import urllib
import pkg_resources
import requests
import functools
import time
from tqdm import tqdm


def timer(func):
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        tic = time.perf_counter()
        value = func(*args, **kwargs)
        toc = time.perf_counter()
        elapsed_time = toc - tic
        print(f"Elapsed time: {elapsed_time:0.4f} seconds")
        return value

    return wrapper_timer


def _single_spaced(x):
    return re.sub(" +", " ", x)


def _local_version(package_name="insightspy"):
    package_name = package_name.lower()
    return next(
        (
            p.version
            for p in pkg_resources.working_set
            if p.project_name.lower() == package_name
        ),
        "No match",
    )


def _versions(package_name="insightspy"):
    url = f"https://pypi.org/pypi/{package_name}/json"
    data = json.load(urllib.request.urlopen(url))
    current_version = _local_version(package_name)
    # Get versions on pip
    versions_list = list(data["releases"].keys())
    versions_list.sort(key=lambda s: list(map(int, s.split("."))))
    latest_version = versions_list[-1]
    if latest_version != current_version:
        print(
            "Local version of insightspy is not the latest version"
            + f"({current_version} < {latest_version})"
        )


def _connect(url, timeout=0.5):
    try:
        r = requests.head(url, timeout=timeout)  # noqa F841
        return True
    except requests.exceptions.ConnectionError:
        return False


def _parse_locus(x):
    return re.split("-|:", x)


def _list_to_int_list(lis):
    out = []
    for x in lis:
        # Convert to float
        try:
            z = float(x)
        except ValueError:
            raise TypeError("one or more elements of the list are not integer valued")
        # check that it is an integer
        if (int(z) - z) == 0:
            out.append(int(z))
        else:
            raise TypeError("one or more elements of the list are not integer valued")
    return out


def _download(urls):
    """Generic downloader

    Downloads files with pretty downloader bar

    Args:
        urls (dict): a dictionary where the keys are the file destinations and the
            values are the urls from which the files are downloaded
    """
    for file, url in urls.items():
        r = requests.get(url, stream=True)
        total_size_in_bytes = int(r.headers.get("content-length", 0))
        block_size = int(1.049e6)  # 1 Mebibyte
        print(f"Downloading {file}...")
        progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
        with open(file, "wb") as f:
            for data in r.iter_content(block_size):
                progress_bar.update(len(data))
                f.write(data)
            progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            raise ValueError("Download did not complete!")



from insightspy.SessionCore import RequestCore
import pandas as pd


class ProjectsMixin(RequestCore):
    def projects(self):
        """List projects

        Lists projects accessible withthin the current portal session. Project access is
        limited by user id.

        Examples:
            >>> # p is a logged in portal session
            >>> p.projects()

        Returns:
            DataFrame: table of project descriptions and ids
        """
        out = [
            {"project_id": v["project_id"], "description": v["description"]}
            for k, v in self._post("project/retrieve", expect_data=True)["response"][
                "data"
            ].items()
        ]
        return pd.DataFrame.from_dict(out)

    def set_project(self, project_id):
        """Set current project

        Limits resources available in the current session to those accessible from the
        specified `project_id`. See available project ids with
        :meth:`~PortalSession.projects`

        Args:
            project_id (int): project id.

        Examples:
            >>> # p is a logged in portal session
            >>> p.set_project(3)
        """
        self._post("project/session", {"project_id": project_id})



import requests
from insightspy.utils import _single_spaced
from abc import ABC


class RequestCore(ABC):
    def __init__(self, api_key=None, url=None):
        session = requests.Session()
        self._session = session
        self.url = url
        if api_key is not None:
            self._update_credentials(api_key)

    def _update_credentials(self, access_token):
        self.access_token = access_token
        self._session.headers.update({"Authorization": "Bearer {access_token}"})

    def _get(self, route, params={}, expect_data=False):
        """Generic GET request

        Wrapper for generic get request to the Arpeggio portal

        Args:
            route (str): string for the route to call. Do not include the core domain
                (e.g. `https://insights.arpeggiobio.com`).
            params (dict): dictionary of parameters to pass to the route
            expect_data (bool): whether to expect data from the response

        Returns:
            result (dict): A dictionary`{status:status_code,
                response:requests.response}`
        """
        request = self._session.get(f"{self.url}/{route}", params=params)
        return RequestCore._clean_response(request, expect_data=expect_data)

    def _post(self, route, json={}, expect_data=False):
        """Generic POST request

        Wrapper for generic POST request to the Arpeggio portal

        Args:
            route (str): string for the route to call. Do not include the core domain
                (e.g. `https://insights.arpeggiobio.com`).
            json (dict): dictionary of parameters to pass to the route in the body of the
                request as a json
            expect_data (bool): whether to expect data from the response

        Returns:
            result (dict): A dictionary `{status:status_code,
                response:requests.response}`
        """
        request = self._session.post(f"{self.url}/{route}", json=json)
        return RequestCore._clean_response(request, expect_data=expect_data)

    @classmethod
    def _clean_response(cls, request, expect_data):
        if request.status_code == 500:
            raise requests.exceptions.RequestException(
                _single_spaced(
                    f"{request.status_code}: server-side error, "
                    + "please contact Arpeggio Biosciences"
                )
            )
        if request.status_code == 404:
            raise requests.exceptions.RequestException(
                _single_spaced(
                    f"{request.status_code}: route does not exist, "
                    + "please contact Arpeggio Biosciences"
                )
            )
        response = request.json()
        if isinstance(response, str):
            response = {"data": response}
        if "notification" not in response:
            response["notification"] = ""
        if request.status_code != 200:
            raise requests.exceptions.RequestException(
                _single_spaced(f'{request.status_code}: {response["notification"]}')
            )
        if expect_data and response["data"] is None:
            raise ValueError(_single_spaced(response["notification"]))
        return {"status": request.status_code, "response": response}



import setuptools
import os
from setuptools.command.install import install
import sys

# get key package details from insightspy/__version__.py
about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "insightspy", "__version__.py")) as f:
    exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
with open("README.md", "r") as f:
    readme = f.read()

# circleci.py version
VERSION = about["__version__"]


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setuptools.setup(
    name=about["__title__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    version=VERSION,
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=setuptools.find_packages(),
    install_requires=["pandas", "tqdm", "requests"],
    include_package_data=True,
    python_requires=">=3.7.*",
    classifiers=about["__classifiers__"],
    cmdclass={
        "verify": VerifyVersionCommand,
    },
)
